/**
 * @file actuator.hpp
 * @author Nawe Robotics
 * @brief File contains IMU class definations
 */

#pragma once
#include "can_device_type.hpp"
#include "interface.hpp"

/**
 * @brief IMU Class
 *
 */
class IMU : public CANDeviceType {

    /**
     * @brief Class defination for IMU Implementation class
     *
     */
    class IMUImpl;

    /**
     * @brief Pointer to class object which stores IMU implementation
     *
     */
    IMUImpl *impl;

    /**
     * @brief CAN Interface object reference. This object is used to send/receive
     * data from CAN Bus.
     *
     */
    Interface *iface;

    enum{
        ENABLE_ROTATION_VECTOR =0x01,
        ENABLE_RAW_ACCELEROMETER,
        ENABLE_RAW_GYRO,
        ENABLE_RAW_MAGNETOMETER,
        ENABLE_LINEAR_ACCELEROMETER,
        ENABLE_ACCELEROMETER,
        ENABLE_GYRO,
        ENABLE_MAGNETOMETER,
        ENABLE_UNCALIBRATED_GYRO,
        ENABLE_GRAVITY,
    };

    /**
     * @brief Callback function which is invoked when a new CAN msg for this
     * device comes into the interface
     *
     * @param fra Reference to the received CAN Msg frame
     */
    void canMsgCallback(struct can_frame_data *fra);
public:
    /**
     * @brief Construct a new IMU object
     *
     * @param device_id Device Id for the IMU device
     * @param interface CAN Interface object for the interface to which the
     * actuator is connected
     */
    IMU(uint32_t device_id, Interface *interface);

    /**
     * @brief Set the Output Log Stream object
     *
     * @param out Output log stream object
     */
    void setOutputLogStream(std::ostream *out);

    ret_status_t enableRotationVector(uint16_t report_interval);

    ret_status_t disableRotationVector();

    ret_status_t enableRawAccelerometer(uint16_t report_interval);

    ret_status_t disableRawAccelerometer();

    ret_status_t enableRawGyro(uint16_t report_interval);

    ret_status_t disableRawGyro();

    ret_status_t enableRawMagnetometer(uint16_t report_interval);

    ret_status_t disableRawMagnetometer();

    ret_status_t enableLinearAccelerometer(uint16_t report_interval);

    ret_status_t disableLinearAccelerometer();

    ret_status_t enableAccelerometer(uint16_t report_interval);

    ret_status_t disableAccelerometer();

    ret_status_t enableGyro(uint16_t report_interval);

    ret_status_t disableGyro();

    ret_status_t enableMagnetometer(uint16_t report_interval);

    ret_status_t disableMagnetometer();

    ret_status_t enableUncalibratedGyro(uint16_t report_interval);

    ret_status_t disableUncalibratedGyro();

    ret_status_t enableGravity(uint16_t report_interval);

    ret_status_t disableGravity();
    /**
     * @brief Function to get Rotation Vector Data of the IMU
     * @param quat_i quaternion i of rotation vector
     * @param quat_j quaternion j of rotation vector
     * @param quat_k quaternion k of rotation vector
     * @param quat_real quaternion real of rotation vector
     * @return Return Status
    */
    ret_status_t getRotationVectorData(float* quat_i, float* quat_j, float* quat_k, float* quat_real);

    /**
     * @brief Function to get Raw Accelerometer Data of the IMU
     * @param x raw accelerometer in x axis
     * @param y raw accelerometer in y axis
     * @param z raw accelerometer in z axis
     * @param accuracy accuracy of the magnetometer data
     * @return Return Status
    */
    ret_status_t getRawAccelerometerData(int16_t* x, int16_t* y, int16_t* z);

    /**
     * @brief Function to get Raw Gyro Data of the IMU
     * @param x raw gyro in x axis
     * @param y raw gyro in y axis
     * @param z raw gyro in z axis
     * @param accuracy accuracy of the magnetometer data
     * @return Return Status
    */
    ret_status_t getRawGyroData(int16_t* x, int16_t* y, int16_t* z);

    /**
     * @brief Function to get Raw Magnetometer Data of the IMU
     * @param x raw magnetometer in x axis
     * @param y raw magnetometer in y axis
     * @param z raw magnetometer in z axis
     * @param accuracy accuracy of the magnetometer data
     * @return Return Status
    */
    ret_status_t getRawMagnetometerData(int16_t* x, int16_t* y, int16_t* z);

    /**
     * @brief Function to get Linear Accelerometer Data of the IMU
     * @param x linear acceleration in x axis in m/s^2
     * @param y linear acceleration in y axis in m/s^2
     * @param z linear acceleration in z axis in m/s^2
     * @param accuracy accuracy of the linear acceleration data
     * @return Return Status
    */
    ret_status_t getLinearAccelerometerData(float* x, float* y, float* z, uint8_t* accuracy);

    /**
     * @brief Function to get Accelerometer Data of the IMU
     * @param x acceleration in x axis in m/s^2
     * @param y acceleration in y axis in m/s^2
     * @param z acceleration in z axis in m/s^2
     * @param accuracy accuracy of the acceleration data
     * @return Return Status
    */
    ret_status_t getAccelerometerData(float* x, float* y, float* z, uint8_t* accuracy);
    
    /**
     * @brief Function to get Gyroscope Data of the IMU
     * @param x gyro in x axis in radians per second
     * @param y gyro in y axis in radians per second
     * @param z gyro in z axis in radians per second
     * @param accuracy accuracy of the gyro data
     * @return Return Status
    */
    ret_status_t getGyroData(float* x, float* y, float* z, uint8_t* accuracy);

    /**
     * @brief Function to get Magnetometer Data of the IMU
     * @param x magnetometer in x axis in uTesla
     * @param y magnetometer in y axis in uTesla
     * @param z magnetometer in z axis in uTesla
     * @param accuracy accuracy of the magnetometer data
     * @return Return Status
    */
    ret_status_t getMagnetometerData(float* x, float* y, float* z, uint8_t* accuracy);

    /**
     * @brief Function to get Uncalibrated Gyroscope Data of the IMU
     * @param x uncalibrated gyro in x axis in radians per second
     * @param y uncalibrated gyro in y axis in radians per second
     * @param z uncalibrated gyro in z axis in radians per second
     * @return Return Status
    */
    ret_status_t getUncalibratedGyroData(float* x, float* y, float* z);

    /**
     * @brief Function to get Uncalibrated Gyroscope Bias Data of the IMU
     * @param x uncalibrated gyro bias in x axis in radians per second
     * @param y uncalibrated gyro bias in y axis in radians per second
     * @param z uncalibrated gyro bias in z axis in radians per second
     * @return Return Status
    */
    ret_status_t getUncalibratedGyroBiasData(float* x, float* y, float* z);

    /**
     * @brief Function to get Data in Euler Angle format
     * @param yaw yaw angle of the imu in degrees
     * @param pitch pitch angle of the imu in degrees
     * @param roll roll angle of the imu in degrees
     * @return Return Status
    */
    ret_status_t getEulerAngles(float* yaw, float* pitch, float* roll);

    /**
     * @brief Function to get Uncalibrated Gyroscope Bias Data of the IMU
     * @param x gravity in x axis in g
     * @param y gravity in  in y axis in g
     * @param z gravity in  in z axis in g
     * @return Return Status
    */
    ret_status_t getGravityData(float* x, float* y, float* z);

    /**
     * @brief Destroy the IMU object
     *
     */
    ~IMU();

    /**
     * @brief Check if the device is still connected or not
     *
     * @return true
     * @return false
     */
    bool isConnected();

    /**
     * @brief Function to flash a program onto the device
     * 
     * @param firmware_path path to firmware file
     * @return ret_status_t 
     */
    ret_status_t flash(std::string firmware_path);
};